// =======================================================================
// ST_INC_UTILS.NSS -- Generic Utility Include
// -----------------------------------------------------------------------
// Generic Utility functions for accomplishing common scripting tasks.
//
// REMEMBER:
// Only put functions here that are totally context independent and
// aren't mainly suitable for a specific kind of task. Put AI-related
// generic functions in "st_inc_ai" and Force-related functions in 
// the "st_inc_force" file.
// -----------------------------------------------------------------------
// Last changed: 2005-08-02
// =======================================================================


// =======================================================================
// FUNCTION PROTOTYPES
// -----------------------------------------------------------------------

// EFFECT MANIPULATION
void   ST_RemoveEffectByType(int nEffectType, object oTarget=OBJECT_SELF);
void   ST_RemoveHarmfulEffects(object oTarget=OBJECT_SELF);
void   ST_RemovePowerEffects(int nPower, object oTarget=OBJECT_SELF);
void   ST_ApplyEffectToHostile(int nDurationType, effect eEffect, object oTarget, float fDur=0.0);
void   ST_ApplyEffectToParty(int nDurationType, effect eEffect, float fDur=0.0);
void   ST_RemoveEffectTypeByCreator(object oTarget, int nEffectType, object oCreator=OBJECT_SELF);
int    ST_GetHasEffect(int nEffectType, object oTarget=OBJECT_SELF);


// MATHS/NUMBERS STUFF
string ST_IntToStringPad(int nVal, int nDigits);
float  ST_FMax(float fValue, float fLimit);
float  ST_FMin(float fValue, float fLimit);
int    ST_IMax(int nValue, int nLimit);
int    ST_IMin(int nValue, int nLimit);


// CREATURE & ITEM MANIPULATION
string ST_ClassToString(int nClass);
void   ST_HealTarget(object oTarget, int bRes=TRUE);
void   ST_FaceObject(object oTarget, object oSource=OBJECT_SELF);
void   ST_SetFacingByObject(object oTarget, object oSource=OBJECT_SELF);
void   ST_SetFacingByLoc(location lLoc, object oSource=OBJECT_SELF);
int    ST_UsesRangedWeapon(object oTarget=OBJECT_SELF);
int    ST_GetHasRocketLauncher(object oTarget=OBJECT_SELF);
int    ST_GetExpForLevel(int nLevel, object oTarget=OBJECT_SELF);
int    ST_GetHasItem(string sTag, object oTarget=OBJECT_SELF);
int    ST_GetIsInFieldOfView(object oTarget, object oSource=OBJECT_SELF, float nFOV=180.0f);


// PARTY MANAGEMENT
void   ST_AddToPartyTable(int nSlot, string sIdent, int bUseTemplate=TRUE, int bDestroy=TRUE);
void   ST_SwitchMainCharacter(int nPlayer, int bRestoreParty=FALSE);
int    ST_GetPartySlotByTag(string sTag);

// STORING LARGER UNSIGNED INTEGERS (16 bit)
void ST_SetLocalShort(object oTarget, int iIdx1, int iIdx2, int iValue);
int ST_GetLocalShort(object oTarget, int iIdx1, int iIdx2);


// FEEDBACK
void   ST_Feedback(string sMessage);
void   ST_DebugMessage(string sText, int bTimestamp=TRUE);


// =======================================================================
// FUNCTION IMPLEMENTATION
// -----------------------------------------------------------------------


//////////////////////////////////////////////////////////////////////////
// EFFECT MANIPULATION - - - - - - - - - - - - - - - - - - - - - - - - - -


// -----------------------------------------------------------------------
// Returns TRUE if oTarget has an effect applied of the specified type. 
// Note that some effects don't have effect types and can't be found.
// -----------------------------------------------------------------------
int ST_GetHasEffect(int nEffectType, object oTarget=OBJECT_SELF) {
    effect eEff = GetFirstEffect(oTarget);
    
    while (GetIsEffectValid(eEff)) {
        if (GetEffectType(eEff) == nEffectType)
            return TRUE;
    
        eEff = GetNextEffect(oTarget);
    }
    return FALSE;
}


// -----------------------------------------------------------------------
// Removes all effects of the specified type from oTarget.
// -----------------------------------------------------------------------
void ST_RemoveEffectByType(int nEffectType, object oTarget=OBJECT_SELF) {
    effect eEff = GetFirstEffect(oTarget);
    
    while (GetIsEffectValid(eEff)) {
        if (GetEffectType(eEff) == nEffectType)
            DelayCommand(0.1, RemoveEffect(oTarget, eEff));
    
        eEff = GetNextEffect(oTarget);
    }   
}


// -----------------------------------------------------------------------
// Removes all (detectable) detrimental effects on oTarget.
// -----------------------------------------------------------------------
void ST_RemoveHarmfulEffects(object oTarget=OBJECT_SELF) {
    effect eEff = GetFirstEffect(oTarget);
    int nType;
    
    while (GetIsEffectValid(eEff)) {
        nType = GetEffectType(eEff);
        if ((nType == EFFECT_TYPE_POISON)   
            || (nType == EFFECT_TYPE_PARALYZE)
            || (nType == EFFECT_TYPE_STUNNED)
            || (nType == EFFECT_TYPE_FRIGHTENED)
            || (nType == EFFECT_TYPE_SLEEP)
            || (nType == EFFECT_TYPE_ABILITY_DECREASE)
            || (nType == EFFECT_TYPE_ATTACK_DECREASE)
            || (nType == EFFECT_TYPE_AC_DECREASE)
            || (nType == EFFECT_TYPE_MOVEMENT_SPEED_DECREASE)
            || (nType == EFFECT_TYPE_SAVING_THROW_DECREASE)
            || (nType == EFFECT_TYPE_DAMAGE_DECREASE)
            || (nType == EFFECT_TYPE_DAMAGE_IMMUNITY_DECREASE)
            || (nType == EFFECT_TYPE_FORCE_RESISTANCE_DECREASE)
            || (nType == EFFECT_TYPE_SKILL_DECREASE)
            || (nType == EFFECT_TYPE_BLINDNESS)
            || (nType == EFFECT_TYPE_ENTANGLE)
            || (nType == EFFECT_TYPE_CONFUSED)
            || (nType == EFFECT_TYPE_DROID_CONFUSED)
            || (nType == EFFECT_TYPE_DROIDSCRAMBLE)
            || (nType == EFFECT_TYPE_MINDTRICK))
        {
            DelayCommand(0.1, RemoveEffect(oTarget, eEff));
        }
        eEff = GetNextEffect(oTarget);
    }
}


// -----------------------------------------------------------------------
// Removes all effects from oTarget that originates from the specified
// force power. The nPower parameter is the line number in Spells.2da 
// that should be looked for.
// -----------------------------------------------------------------------
void ST_RemovePowerEffects(int nPower, object oTarget=OBJECT_SELF) {
    if (GetHasSpellEffect(nPower, oTarget)) {
        effect eEff = GetFirstEffect(oTarget);
        while (GetIsEffectValid(eEff)) {
            if( GetEffectSpellId(eEff) == nPower) {
                RemoveEffect(oTarget, eEff);
            }
            eEff = GetNextEffect(oTarget);
        }
    }   
}


// -----------------------------------------------------------------------
// Applies an effect to the target is the target is hostile only. Use 
// this when applying effects with a delay to make sure the target is
// still hostile.
// -----------------------------------------------------------------------
void ST_ApplyEffectToHostile(int nDurationType, effect eEffect, object oTarget, float fDur=0.0) {
    if(GetIsObjectValid(oTarget) && !GetIsDead(oTarget) && GetIsEnemy(oTarget)) {
        ApplyEffectToObject(nDurationType, eEffect, oTarget, fDur);
    }
}


// -----------------------------------------------------------------------
// Applies the specified effect to all members of the active player party.
// -----------------------------------------------------------------------
void ST_ApplyEffectToParty(int nDurationType, effect eEffect, float fDur=0.0) {
    int i;
    for (i = 0; i < GetPartyMemberCount(); i++)
        ApplyEffectToObject(nDurationType, eEffect, GetPartyMemberByIndex(i), fDur);    
}


// -----------------------------------------------------------------------
// Remove all effects from oTarget that are of specified type and are
// created by oCreator.
// -----------------------------------------------------------------------
void ST_RemoveEffectTypeByCreator(object oTarget, int nEffectType, object oCreator=OBJECT_SELF) {
    effect eEff = GetFirstEffect(oTarget);

    while(GetIsEffectValid(eEff)) {
        if ((GetEffectCreator(eEff) == oCreator) && (GetEffectType(eEff) == nEffectType)) {
            RemoveEffect(oTarget, eEff);
        }
        eEff = GetNextEffect(oTarget);
    }
}


//////////////////////////////////////////////////////////////////////////
// CREATURE & ITEMS  - - - - - - - - - - - - - - - - - - - - - - - - - - -


// -----------------------------------------------------------------------
// Returns TRUE if oTarget has a Wrist Launcher item equipped.
// -----------------------------------------------------------------------
int ST_GetHasRocketLauncher(object oTarget=OBJECT_SELF) {
    object oItem = GetItemInSlot( INVENTORY_SLOT_LEFTARM, oTarget );
    if( GetIsObjectValid(oItem) && (GetBaseItemType(oItem) == BASE_ITEM_WRIST_LAUNCHER)) {
        return TRUE;
    }
    else {
        oItem = GetItemInSlot( INVENTORY_SLOT_RIGHTARM, oTarget );
        if( GetIsObjectValid(oItem) && (GetBaseItemType(oItem) == BASE_ITEM_WRIST_LAUNCHER)) {
            return TRUE;
        }
    }
    return FALSE;
}


// -----------------------------------------------------------------------
// Returns TRUE if oTarget has a ranged weapon equipped.
// -----------------------------------------------------------------------
int ST_UsesRangedWeapon(object oTarget=OBJECT_SELF) {
    object oItem = GetItemInSlot(INVENTORY_SLOT_RIGHTWEAPON, oTarget);

    if(GetIsObjectValid(oItem)) {
        switch (GetBaseItemType(oItem)) {
            case BASE_ITEM_BLASTER_CARBINE:
            case BASE_ITEM_BLASTER_PISTOL:
            case BASE_ITEM_BLASTER_RIFLE:
            case BASE_ITEM_HEAVY_BLASTER:
            case BASE_ITEM_HEAVY_REPEATING_BLASTER:
            case BASE_ITEM_REPEATING_BLASTER:
            case BASE_ITEM_HOLD_OUT_BLASTER:
            case BASE_ITEM_BOWCASTER:       
            case BASE_ITEM_DISRUPTER_PISTOL:
            case BASE_ITEM_DISRUPTER_RIFLE:
            case BASE_ITEM_ION_BLASTER:
            case BASE_ITEM_ION_RIFLE:
            case BASE_ITEM_SONIC_PISTOL:
            case BASE_ITEM_SONIC_RIFLE:  return TRUE; break;        
        }
    }
    return FALSE;
}


// -----------------------------------------------------------------------
// Heals the targets Vitality and Force points, and removes (most) 
// detrimental effects from it. Unless the bRes parameter is set to FALSE
// the target will be resurrected as well if they are dead.
// -----------------------------------------------------------------------
void ST_HealTarget(object oTarget, int bRes=TRUE) {
    if (bRes)
        ApplyEffectToObject(DURATION_TYPE_INSTANT, EffectResurrection(), oTarget);
        
    ApplyEffectToObject(DURATION_TYPE_INSTANT, EffectHeal(GetMaxHitPoints(oTarget)), oTarget);
    ApplyEffectToObject(DURATION_TYPE_INSTANT, EffectHealForcePoints(GetMaxForcePoints(oTarget)), oTarget);
    ApplyEffectToObject(DURATION_TYPE_INSTANT, EffectVisualEffect(VFX_IMP_HEAL), oTarget);
    
    ST_RemoveHarmfulEffects(oTarget);
}


// -----------------------------------------------------------------------
// Returns the amount of experience points needed by oTarget to reach
// the specified character level.
// -----------------------------------------------------------------------
int  ST_GetExpForLevel(int nLevel, object oTarget=OBJECT_SELF) {
    return (nLevel * (nLevel - 1) * 500) - GetXP(oTarget);
}


// -----------------------------------------------------------------------
// Checks if oTarget has an item with the tag sTag in their inventory.
// Set oTarget to the main character (GetFirstPC) to check if the item is
// in the party inventory.
// -----------------------------------------------------------------------
int ST_GetHasItem(string sTag, object oTarget=OBJECT_SELF) {
    return GetIsObjectValid(GetItemPossessedBy(oTarget, sTag));
}


// -----------------------------------------------------------------------
// Makes oSource turn to face oTarget. oTarget can be any type of object
// with a facing (waypoints, placeables, creatures...).
// -----------------------------------------------------------------------
void ST_FaceObject(object oTarget, object oSource=OBJECT_SELF) {
    AssignCommand(oSource, SetFacingPoint(GetPosition(oTarget)));
}


// -----------------------------------------------------------------------
// Makes oSource turn to face in the same direction as oTarget. oTarget
// can be any type of object with a facing, such as waypoints, placeables, 
// creatures...
// -----------------------------------------------------------------------
void ST_SetFacingByObject(object oTarget, object oSource=OBJECT_SELF) {
    AssignCommand(oSource, SetFacing(GetFacing(oTarget)));
}


// -----------------------------------------------------------------------
// Makes oSource turn to face in the direction indicated by lLoc.
// -----------------------------------------------------------------------
void ST_SetFacingByLoc(location lLoc, object oSource=OBJECT_SELF) {
    AssignCommand(oSource, SetFacing(GetFacingFromLocation(lLoc)));
}


// -----------------------------------------------------------------------
// Returns the name, as a string, of the class specified by nClass, which
// is one of the CLASS_TYPE_* constants representing a row in classes.2da.
// -----------------------------------------------------------------------
string ST_ClassToString(int nClass) {
    string sClass;
    switch (nClass) {
        case CLASS_TYPE_SOLDIER:            sClass = "Soldier";             break;
        case CLASS_TYPE_SCOUT:              sClass = "Scout";               break;
        case CLASS_TYPE_SCOUNDREL:          sClass = "Scoundrel";           break;
        case CLASS_TYPE_JEDIGUARDIAN:       sClass = "Jedi Guardian";       break;
        case CLASS_TYPE_JEDICONSULAR:       sClass = "Jedi Consular";       break;
        case CLASS_TYPE_JEDISENTINEL:       sClass = "Jedi Sentinel";       break;
        case CLASS_TYPE_COMBATDROID:        sClass = "Combat Droid";        break;
        case CLASS_TYPE_EXPERTDROID:        sClass = "Expert Droid";        break;
        case CLASS_TYPE_MINION:             sClass = "Minion";              break;
        case CLASS_TYPE_TECHSPECIALIST:     sClass = "Tech Specialist";     break;
        case CLASS_TYPE_BOUNTYHUNTER:       sClass = "Bounty Hunter";       break;
        case CLASS_TYPE_JEDIWEAPONMASTER:   sClass = "Jedi Weapon Master";  break;
        case CLASS_TYPE_JEDIMASTER:         sClass = "Jedi Master";         break;
        case CLASS_TYPE_JEDIWATCHMAN:       sClass = "Jedi Watchman";       break;
        case CLASS_TYPE_SITHMARAUDER:       sClass = "Sith Marauder";       break;
        case CLASS_TYPE_SITHLORD:           sClass = "Sith Lord";           break;
        case CLASS_TYPE_SITHASSASSIN:       sClass = "Sith Assassin";       break;
        default:                            sClass = "Unknown Class";       break;
    }
    return sClass;
}


// -----------------------------------------------------------------------
// Returns TRUE if oTarget is in front of oSource.
// -----------------------------------------------------------------------
int ST_GetIsInFieldOfView(object oTarget, object oSource=OBJECT_SELF, float nFOV=180.0f) {
    if (!GetIsObjectValid(oTarget) || !GetIsObjectValid(oSource))
        return FALSE;

    if (!GetObjectSeen(oTarget, oSource))
        return FALSE;

    vector vMPos = GetPosition(oTarget) - GetPosition(oSource);
    vector vOff  = AngleToVector(GetFacing(oSource));

    vMPos = VectorNormalize(vMPos);
    vOff  = VectorNormalize(vOff);

    float fAngle = acos((vMPos.x * vOff.x) + (vMPos.y * vOff.y) + (vMPos.z * vOff.z));

    return (fAngle < (nFOV / 2.0f));
}


//////////////////////////////////////////////////////////////////////////
// PARTY MANAGEMENT  - - - - - - - - - - - - - - - - - - - - - - - - - - - 


// -----------------------------------------------------------------------
// Get the party table slot that a NPC with the specified tag usually
// goes in. Only supports the tags of the standard party members.
// -----------------------------------------------------------------------
int ST_GetPartySlotByTag(string sTag) {
    if (sTag == "atton")            return NPC_ATTON;
    else if (sTag == "baodur")      return NPC_BAO_DUR;
    else if (sTag == "mand")        return NPC_CANDEROUS;
    else if (sTag == "disciple")    return NPC_DISCIPLE;
    else if (sTag == "g0t0")        return NPC_G0T0;
    else if (sTag == "handmaiden")  return NPC_HANDMAIDEN;
    else if (sTag == "hanharr")     return NPC_HANHARR;
    else if (sTag == "hk47")        return NPC_HK_47;
    else if (sTag == "kreia")       return NPC_KREIA;
    else if (sTag == "mira")        return NPC_MIRA;
    else if (sTag == "t3m4")        return NPC_T3_M4;
    else if (sTag == "visasmarr")   return NPC_VISAS;
    
    return 99;
}


// -----------------------------------------------------------------------
// Removes the Exile from the party and turn the character in the nPlayer
// party table slot into the main character. If bRestoreParty is set to
// TRUE, the other party members will be put back into the party along
// with the new main char. If left to False, the new main char will be
// alone.
// -----------------------------------------------------------------------
void ST_SwitchMainCharacter(int nPlayer, int bRestoreParty=FALSE) {
    int nNPC;
    int nParty1 = -1;
    int nParty2 = -1;
    
    for (nNPC = NPC_ATTON; nNPC <= NPC_DISCIPLE; nNPC++) {  
        if (IsNPCPartyMember(nNPC)) {
            if (bRestoreParty 
                && (nParty1 == -1) 
                && (nNPC != nPlayer))
            {
                nParty1 = nNPC;
            }
            else if (bRestoreParty 
                && (nParty2 == -1) 
                && (nNPC != nPlayer))
            {
                nParty2 = nNPC;
            }           
                
            RemoveNPCFromPartyToBase(nNPC);
            
        }
    }
    
    SwitchPlayerCharacter(nPlayer);
    
    if (bRestoreParty && (nParty1 != -1)) {
        object oNPC = SpawnAvailableNPC(nParty1, GetLocation(GetFirstPC()));    
        AddPartyMember(nParty1, oNPC);
    }
    if (bRestoreParty && (nParty2 != -1)) {
        object oNPC = SpawnAvailableNPC(nParty2, GetLocation(GetFirstPC()));    
        AddPartyMember(nParty2, oNPC);
    }   
}


// -----------------------------------------------------------------------
// Add a new NPC to the party table in the specified slot. If bUseTemplate
// is TRUE, sIdent is the ResRef of the NPC to add. If bUseTemplate is
// false, sIdent is the Tag of a creature in the area to add. 
// If the specified party table slot is already occupied by another NPC,
// that NPC will be removed. If the NPC is in the active party it will
// be removed from the party and the area, unless bDestroy is set to
// FALSE, in which case the creature will remain standing in the area
// after being removed from the party.
// -----------------------------------------------------------------------
void ST_AddToPartyTable(int nSlot, string sIdent, int bUseTemplate=TRUE, int bDestroy=TRUE) {
    if (IsNPCPartyMember(nSlot)) {
        if (bDestroy)
            RemoveNPCFromPartyToBase(nSlot);
        else
            RemovePartyMember(nSlot);   
    }
    
    if (IsAvailableCreature(nSlot))
        RemoveAvailableNPC(nSlot);
        
    if (bUseTemplate)
        AddAvailableNPCByTemplate(nSlot, sIdent);
    else
        AddAvailableNPCByObject(nSlot, GetNearestObjectByTag(sIdent, GetFirstPC()));
        
    if (GetPartyMemberCount() < 3) {
        object oNPC = SpawnAvailableNPC(nSlot, GetLocation(GetFirstPC()));
        AddPartyMember(nSlot, oNPC);
    }
}


//////////////////////////////////////////////////////////////////////////
// NUMBER MANIPULATION - - - - - - - - - - - - - - - - - - - - - - - - - -


// -----------------------------------------------------------------------
// Returns the highest value of the two floating point values.
// -----------------------------------------------------------------------
float ST_FMax(float fValue, float fLimit) {
    if (fValue < fLimit)
        fValue = fLimit;
        
    return fValue;  
}


// -----------------------------------------------------------------------
// Returns the highest value of the two integer values.
// -----------------------------------------------------------------------
int ST_IMax(int nValue, int nLimit) {
    if (nValue < nLimit)
        nValue = nLimit;
        
    return nValue;  
}


// -----------------------------------------------------------------------
// Returns the lowest value of the two floating point values.
// -----------------------------------------------------------------------
float ST_FMin(float fValue, float fLimit) {
    if (fValue > fLimit)
        fValue = fLimit;
        
    return fValue;  
}


// -----------------------------------------------------------------------
// Returns the lowest value of the two integer values.
// -----------------------------------------------------------------------
int ST_IMin(int nValue, int nLimit) {
    if (nValue > nLimit)
        nValue = nLimit;
        
    return nValue;  
}

//////////////////////////////////////////////////////////////////////////
// STORING LARGER UNSIGNED INTEGERS - - - - - - - - - - - - - - - - - - -

void ST_SetLocalShort(object oTarget, int iIdx1, int iIdx2, int iValue) {
	int iP1 = iValue & 0x000000FF;
	int iP2 = (iValue >> 8) & 0x000000FF;
	
	iP1 = (iP1 > 127 ? -128 + (iP1 - 128) : iP1);
	iP2 = (iP2 > 127 ? -128 + (iP2 - 128) : iP2);
	
	SetLocalNumber(oTarget, iIdx1, iP1);
	SetLocalNumber(oTarget, iIdx2, iP2);	
}


int ST_GetLocalShort(object oTarget, int iIdx1, int iIdx2) {
	int iP1 = GetLocalNumber(OBJECT_SELF, 20);
	int iP2 = GetLocalNumber(OBJECT_SELF, 21);
	
	return (iP1 | (iP2 << 8));
}


// -----------------------------------------------------------------------
// Return string value of the integer nVal zero-padded to the number of 
// digits specified by the nDigits parameter.
// -----------------------------------------------------------------------
string ST_IntToStringPad(int nVal, int nDigits) {
    string sRes = IntToString(nVal);
    if (sRes != "") {
        int nDiff = nDigits - GetStringLength(sRes);
        
        if (nDiff > 0) {
            int i;
            for (i = 0; i < nDiff; i++) {
                sRes = "0" + sRes;  
            }   
        }
    }
    return sRes;
}


//////////////////////////////////////////////////////////////////////////
// FEEDBACK FUNCTIONS  - - - - - - - - - - - - - - - - - - - - - - - - - - 


// -----------------------------------------------------------------------
// Simple wrapper to dump a text message to the Feedback screen.
// -----------------------------------------------------------------------
void ST_Feedback(string sMessage) {
    SendMessageToPC(GetPartyLeader(), sMessage);
}


// -----------------------------------------------------------------------
// Writes a time-stamped debug message to the Feedback log, 
// indicating the name of the object that ran the function. Set the 
// bTimestamp parameter to FALSE to remove the timestamp.
// -----------------------------------------------------------------------
void ST_DebugMessage(string sText, int bTimestamp=TRUE) {
    string sTimestamp = "";
    
    if (bTimestamp) {
        int nHour = GetTimeHour();
        int nMin  = GetTimeMinute(); // ST: Ingame-time! There are 2 min/hour.
        int nSec  = GetTimeSecond();
        int nMSec = GetTimeMillisecond();
        
        sTimestamp = (nHour < 10 ? "0" : "") + IntToString(nHour)
             + ":" + (nMin  < 10 ? "0" : "") + IntToString(nMin)
             + ":" + (nSec  < 10 ? "0" : "") + IntToString(nSec)
             + ":" + GetSubString((nMSec  < 10 ? "0" : "") + IntToString(nMSec), 0, 2); 
             
        sTimestamp += " - ";    
    }
    
    SendMessageToPC(GetPartyLeader(), sTimestamp + "DEBUG[" + GetName(OBJECT_SELF) + "] " + sText);
}
